from __future__ import annotations

import copy
from enum import StrEnum
from typing import Annotated, Any, Dict, List

import httpx
from pydantic import BaseModel, BeforeValidator, Field, PlainSerializer, model_validator

from .resource_abc import CamelAlias, Ref, Resource, register_resource


class TypeValueRange(StrEnum):
    OPEN = "Open"
    CLOSED = "Closed"


class ValueType(StrEnum):
    BOOLEAN = "Boolean"
    CODE = "Code"
    CURRENCY = "Currency"
    CURRENCY_AND_AMOUNT = "CurrencyAndAmount"
    CUT_LOCAL_TIME = "CutLocalTime"
    DATE_OR_CUTLABEL = "DateOrCutLabel"
    DATE_TIME = "DateTime"
    DECIMAL = "Decimal"
    ID = "Id"
    INT = "Int"
    LIST = "List"
    MAP = "Map"
    METRIC_VALUE = "MetricValue"
    PERCENTAGE = "Percentage"
    PROPERTY_ARRAY = "PropertyArray"
    RESOURCE_ID = "ResourceId"
    RESULT_VALUE = "ResultValue"
    STRING = "String"
    TRADE_PRICE = "TradePrice"
    UNINDEXED_TEXT = "UnindexedText"
    URI = "Uri"


class FieldValueType(StrEnum):
    STRING = "String"
    DECIMAL = "Decimal"


class Unit(BaseModel, CamelAlias):
    code: str
    display_name: str
    description: str
    details: Any | None = None


class UnitSchema(StrEnum):
    NO_UNITS = "NoUnits"
    BASIC = "Basic"
    ISO4217_CURRENCY = "Iso4217Currency"


class FieldDefinition(BaseModel, CamelAlias):
    key: str
    is_required: bool
    is_unique: bool
    value_type: str | FieldValueType = FieldValueType.STRING


class FieldValue(BaseModel):
    value: str
    fields: Dict[str, str]  # limit to strings because get returns strings


class ReferenceData(BaseModel, CamelAlias):
    field_definitions: List[FieldDefinition]
    values: List[FieldValue]


@register_resource()
class DataTypeRef(BaseModel, Ref):
    id: str = Field(exclude=True)
    scope: str
    code: str

    def attach(self, client):
        scope, code = self.scope, self.code
        try:
            client.get(f"/api/api/datatypes/{scope}/{code}", params={"includeSystem": True})
        except httpx.HTTPStatusError as ex:
            if ex.response.status_code == 404:
                raise RuntimeError(f"Datatype {scope}/{code} not found")
            else:
                raise ex


@register_resource()
class DataTypeResource(CamelAlias, BaseModel, Resource):
    id: str = Field(exclude=True)
    scope: str
    code: str
    type_value_range: TypeValueRange
    display_name: str
    description: str
    value_type: ValueType
    acceptable_values: List[str] | None = None
    unit_schema: UnitSchema = UnitSchema.NO_UNITS
    acceptable_units: List[Unit] | None = None
    reference_data: ReferenceData | None = None

    @model_validator(mode="before")
    @classmethod
    def des_model(cls, data, info) -> Dict[str, Any]:
        if not isinstance(data, dict):
            return data
        if isinstance(data.get("id", None), dict):
            data = data | data.pop("id")
        if info.context and info.context.get("id"):
            return data | {"id": info.context.get("id")}
        return data

    def read(self, client, old_state) -> Dict[str, Any]:
        return client.get(f"/api/api/datatypes/{old_state.scope}/{old_state.code}").json()

    def create(self, client: httpx.Client):
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True)
        client.post("/api/api/datatypes", json=desired)
        return {"scope": self.scope, "code": self.code}

    def update(self, client: httpx.Client, old_state):
        if [self.scope, self.code] != [old_state.scope, old_state.code]:
            raise RuntimeError("Cannot change scope/code on datatype")
        remote = self.read(client, old_state)
        remote.pop("href")
        remote.pop("id")
        remote.pop("links")
        remote.pop("version")
        desired = self.model_dump(
            mode="json", exclude_none=True, exclude={"scope", "code"}, by_alias=True
        )
        effective = remote | copy.deepcopy(desired)
        if "referenceData" in remote and "referenceData" in effective:
            # sort the fields since the api changes the order they come back
            remote["referenceData"]["fieldDefinitions"].sort(key=lambda field: field["key"])
            effective["referenceData"]["fieldDefinitions"].sort(key=lambda field: field["key"])
            # sort the values since the api changes the order they come back
            remote["referenceData"]["values"].sort(key=lambda field: field["value"])
            effective["referenceData"]["values"].sort(key=lambda field: field["value"])
        if "acceptableValues" in remote and "acceptableValues" in effective:
            # sort the acceptableValues since the api changes the order they come back
            remote["acceptableValues"].sort()
            effective["acceptableValues"].sort()
        if effective == remote:
            return None
        # check for illegal modifications
        readonly_fields = ["typeValueRange", "unitSchema", "valueType"]
        modified = [field for field in readonly_fields if effective[field] != remote[field]]
        if len(modified) > 0:
            raise RuntimeError(f"Cannot change readonly fields {modified} on datatype")
        if effective["referenceData"]["fieldDefinitions"] != remote["referenceData"]["fieldDefinitions"]:
            raise RuntimeError(
                "Cannot change readonly fields referenceData.fieldDefinitions on datatype"
            )
        # update reference data values if required
        if effective["referenceData"]["values"] != remote["referenceData"]["values"]:
            client.put(
                f"/api/api/datatypes/{self.scope}/{self.code}/referencedatavalues",
                json=effective["referenceData"]["values"],
            )
        # update core data if required
        effective.pop("referenceData")
        remote.pop("referenceData")
        if effective != remote:
            client.put(f"/api/api/datatypes/{self.scope}/{self.code}", json=desired)
        return {"scope": self.scope, "code": self.code}

    @staticmethod
    def delete(client, old_state):
        client.delete(f"/api/api/datatypes/{old_state.scope}/{old_state.code}")

    def deps(self):
        return []


class ResourceId(BaseModel):
    scope: str
    code: str


def ser_datatype_key(value: ResourceId | DataTypeRef | DataTypeResource, info):
    style = info.context.get("style", "api") if info.context else "api"
    if style == "dump":
        if isinstance(value, (Resource, Ref)):
            return {"$ref": value.id}
        return value
    # uses str.format to interpolate the values for the api request
    return {"scope": value.scope, "code": value.code}


def des_datatype_key(value, info):
    if info.context and info.context.get("$refs"):
        return info.context["$refs"][value["$ref"]]
    return value


# as a special case, we allow system datatypes to be represented as ResourceId
# so users don't have to create refs for them.
DataTypeKey = Annotated[
    ResourceId | DataTypeRef | DataTypeResource,
    BeforeValidator(des_datatype_key), PlainSerializer(ser_datatype_key)
]



from __future__ import annotations

from enum import StrEnum
from types import SimpleNamespace
from typing import Annotated, Any, Dict, List, TypeVar

import httpx
from pydantic import (
    BaseModel,
    BeforeValidator,
    Field,
    PlainSerializer,
    field_serializer,
    field_validator,
    model_validator,
)

from fbnconfig import datatype

from .resource_abc import CamelAlias, Ref, Resource, register_resource

# Converts a Resource or Ref into an embedded
# reference to the Resource for use when dumping
T = TypeVar("T")


def as_reference(v: T) -> Dict | T:
    if isinstance(v, (Resource, Ref)):
        return {"$ref": v.id}
    return v


ResourceId = datatype.ResourceId  # expose property.ResourceId for backwards compatibility


class LifeTime(StrEnum):
    Perpetual = "Perpetual"
    TimeVariant = "TimeVariant"


class Domain(StrEnum):
    Abor = "Abor"
    AborConfiguration = "AborConfiguration"
    AccessMetadata = "AccessMetadata"
    Account = "Account"
    Allocation = "Allocation"
    Analytic = "Analytic"
    Block = "Block"
    Calendar = "Calendar"
    ChartOfAccounts = "ChartOfAccounts"
    Compliance = "Compliance"
    ConfigurationRecipe = "ConfigurationRecipe"
    CustodianAccount = "CustodianAccount"
    CustomEntity = "CustomEntity"
    CutLabelDefinition = "CutLabelDefinition"
    DerivedValuation = "DerivedValuation"
    DiaryEntry = "DiaryEntry"
    Execution = "Execution"
    FundConfiguration = "FundConfiguration"
    Fund = "Fund"
    Holding = "Holding"
    IdentifierDefinition = "IdentifierDefinition"
    Instrument = "Instrument"
    InstrumentEvent = "InstrumentEvent"
    InvestorRecord = "InvestorRecord"
    InvestmentAccount = "InvestmentAccount"
    Leg = "Leg"
    LegalEntity = "LegalEntity"
    MarketData = "MarketData"
    NextBestAction = "NextBestAction"
    NotDefined = "NotDefined"
    Order = "Order"
    OrderInstruction = "OrderInstruction"
    Package = "Package"
    Participation = "Participation"
    Person = "Person"
    Placement = "Placement"
    Portfolio = "Portfolio"
    PortfolioGroup = "PortfolioGroup"
    PropertyDefinition = "PropertyDefinition"
    Reconciliation = "Reconciliation"
    ReferenceHolding = "ReferenceHolding"
    Transaction = "Transaction"
    TransactionConfiguration = "TransactionConfiguration"
    UnitResult = "UnitResult"


class ConstraintStyle(StrEnum):
    Property = "Property"
    Collection = "Collection"
    Identifier = "Identifier"


class CollectionType(StrEnum):
    Set = "Set"
    Array = "Array"


@register_resource()
class DefinitionRef(BaseModel, Ref):
    id: str = Field(exclude=True)
    domain: Domain
    scope: str
    code: str

    def __format__(self, _):
        return f"Properties[{self.domain}/{self.scope}/{self.code}]"

    def attach(self, client):
        domain, scope, code = self.domain, self.scope, self.code
        try:
            client.get(f"/api/api/propertydefinitions/{domain}/{scope}/{code}")
        except httpx.HTTPStatusError as ex:
            if ex.response.status_code == 404:
                raise RuntimeError(f"Property definition {domain}/{scope}/{code} not found")
            else:
                raise ex


class Formula(BaseModel):
    formula: str
    args: Dict

    def __init__(self, formula, **kwargs):
        super().__init__(formula=formula, args=kwargs)

    @classmethod
    def create(cls, formula, **kwargs):
        return Formula(formula, **kwargs)


@register_resource()
class DefinitionResource(CamelAlias, BaseModel, Resource):
    id: str = Field(init=True, exclude=True)
    domain: Domain
    scope: str
    code: str
    display_name: str
    data_type_id: datatype.DataTypeKey
    property_description: str | None = None
    life_time: LifeTime | None = None
    constraint_style: ConstraintStyle | None = None
    collection_type: str | None = None
    derivation_formula: Formula | None = None
    is_filterable: bool | None = None
    remote: Dict[str, Any] | None = Field(None, exclude=True, init=False)

    def __format__(self, _):
        return f"Properties[{self.domain}/{self.scope}/{self.code}]"

    @field_serializer("derivation_formula", when_used="always")
    def ser_formula(self, value: Formula, info) -> str | Dict[str, Any]:
        style = info.context.get("style", "api") if info.context else "api"
        if style == "dump":
            return {
                "formula": value.formula,
                "args": {k: as_reference(v) for k, v in value.args.items()},
            }
        # use str.format to interpolate the values, hits __format__ when
        # arg is another property
        return value.formula.format(**value.args)

    @field_validator("derivation_formula", mode="before")
    @classmethod
    def des_formula(cls, value, info) -> Formula | None:
        if isinstance(value, Formula):
            return value
        args = {
            k: info.context["$refs"][p["$ref"]] if isinstance(p, Dict) and p.get("$ref") else p
            for k, p in value["args"].items()
        } if info.context and info.context.get("$refs") else value["args"]
        return Formula(value["formula"], **args)

    @model_validator(mode="before")
    @classmethod
    def des_model(cls, data: Any, info) -> Dict[str, Any]:
        if not isinstance(data, dict):
            return data
        if info.context and info.context.get("id"):
            return data | {"id": info.context.get("id")}
        return data

    @model_validator(mode="after")
    def check_derived_or_plain(self):
        if self.derivation_formula:
            if self.life_time or self.constraint_style or self.collection_type:
                raise RuntimeError("A property must be either derived or plain")
        # if this is not a derived property, then is_filterable cannot be set
        elif self.is_filterable is not None:
            raise RuntimeError(
                "Cannot set 'is_filterable' field, a property must be either derived or plain"
            )
        return self

    def read(self, client, old_state):
        domain = old_state.domain
        scope = old_state.scope
        code = old_state.code
        self.remote = client.get(f"/api/api/propertydefinitions/{domain}/{scope}/{code}").json()
        return self.remote

    def create(self, client: httpx.Client):
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True)
        derived = self.derivation_formula is not None
        if derived:
            client.post("/api/api/propertydefinitions/derived", json=desired)
        else:
            client.post("/api/api/propertydefinitions", json=desired)
        return {"domain": self.domain, "scope": self.scope, "code": self.code, "derived": derived}

    @staticmethod
    def delete(client, old_state: SimpleNamespace):
        domain, scope, code = old_state.domain, old_state.scope, old_state.code
        client.delete(f"/api/api/propertydefinitions/{domain}/{scope}/{code}")

    def update(self, client, old_state):
        # cannot change identifier or switch derived and non-derived. recreate
        derived = self.derivation_formula is not None
        if [self.domain, self.scope, self.code, derived] != [
            old_state.domain,
            old_state.scope,
            old_state.code,
            old_state.derived,
        ]:
            self.delete(client, old_state)
            return self.create(client)
        self.read(client, old_state)
        remote = self.remote or {}
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True)
        effective = remote | desired
        if effective == remote:
            return None
        # cannot change dataType, or collectionType need to recreate
        if remote["dataTypeId"] != desired["dataTypeId"] or remote.get(
            "collectionType", None
        ) != effective.get("collectionType", None):
            self.delete(client, old_state)
            return self.create(client)
        if derived:
            client.put(
                f"/api/api/propertydefinitions/derived/{self.domain}/{self.scope}/{self.code}",
                json=desired,
            )
        else:
            client.put(
                f"/api/api/propertydefinitions/{self.domain}/{self.scope}/{self.code}", json=desired
            )
        return {"domain": self.domain, "scope": self.scope, "code": self.code, "derived": derived}

    def deps(self):
        res: List[Resource | Ref] = []
        if self.derivation_formula:
            res = [
                value
                for value in self.derivation_formula.args.values()
                if isinstance(value, (DefinitionResource, DefinitionRef))
            ]
        if isinstance(self.data_type_id, (Resource, Ref)):
            res.append(self.data_type_id)
        return res


#
# for use when another object needs to reference a property definition
# normal form is domain/scope/code
#
def ser_property_key(value, info):
    if info.context and info.context.get("style") == "dump":
        return {"$ref": value.id}
    return "/".join([value.domain, value.scope, value.code])


def des_property_key(value, info):
    if not isinstance(value, dict):
        return value
    if info.context and info.context.get("$refs"):
        ref = info.context["$refs"][value["$ref"]]
        return ref
    return value


PropertyKey = Annotated[
    DefinitionResource | DefinitionRef,
    BeforeValidator(des_property_key), PlainSerializer(ser_property_key)
]



from fbnconfig import Deployment, customentity

"""
An example configuration for a custom entity.
The script configures the following entities:
- Entity Type

More information can be found here:
https://support.lusid.com/knowledgebase/article/KA-01750/en-us
"""


def configure(env):
    ce = customentity.EntityTypeResource(
        id="ce1",
        entity_type_name="entity-type-name",
        display_name="Example Custom Entity",
        description="An example custom entity",
        field_schema=[
            customentity.FieldDefinition(
                name="Field1",
                lifetime=customentity.LifeTime.PERPETUAL,
                type=customentity.FieldType.STRING,
                collection_type=customentity.CollectionType.SINGLE,
                required=True,
            ),
            customentity.FieldDefinition(
                name="Field2",
                lifetime=customentity.LifeTime.TIMEVARIANT,
                type=customentity.FieldType.STRING,
                required=True,
            ),
        ],
    )
    return Deployment("custom_entity_example", [ce])



import os
import time
from types import SimpleNamespace

from pytest import fixture

import fbnconfig
from fbnconfig import lumi
from fbnconfig import property as prop
from tests.integration.generate_test_name import gen

if os.getenv("FBN_ACCESS_TOKEN") is None or os.getenv("LUSID_ENV") is None:
    raise (
        RuntimeError("Both FBN_ACCESS_TOKEN and LUSID_ENV variables need to be set to run these tests")
    )

lusid_env = os.environ["LUSID_ENV"]
token = os.environ["FBN_ACCESS_TOKEN"]
host_vars = {}
client = fbnconfig.create_client(lusid_env, token)


def put_query(client, sql):
    res = client.put(
        "/honeycomb/api/Sql/json",
        content=sql,
        params={"jsonProper": True},
        headers={"Content-type": "text/plain"},
    )
    return res.json()


@fixture(scope="module")
def setup_deployment():
    deployment_name = gen("inlineproperties")
    print(f"\nRunning for deployment {deployment_name}...")
    yield SimpleNamespace(name=deployment_name)
    # Teardown: Clean up resources (if any) after the test
    print("\nTearing down resources...")
    fbnconfig.deploy(fbnconfig.Deployment(deployment_name, []), lusid_env, token)


@fixture()
def get_deployment(setup_deployment):
    # Reference an existing property definition
    test_property_one = prop.DefinitionRef(
        id="test-property-ref", domain=prop.Domain.Transaction, scope="default", code="Custodian"
    )
    test_property_two = prop.DefinitionRef(
        id="test-property-two", domain=prop.Domain.Transaction, scope="default",
        code="SourcePortfolioScope"
    )
    # Create inline properties resource
    inline_properties = lumi.InlinePropertiesResource(
        id="test-inline-properties",
        provider="Lusid.Portfolio.Txn",
        provider_name_extension=setup_deployment.name,
        properties=[
            lumi.InlineProperty(
                key=test_property_one,
                name="TestProperty",
                data_type=lumi.InlineDataType.Text,
                description="Test property for integration testing",
            ),
            lumi.InlineProperty(
                key=test_property_two,
                name="TestProperty2",
                data_type=lumi.InlineDataType.Text,
                description="Test property for integration testing",
            )
        ],
    )
    # create custom entity inline for the fbnconfig setup log (because we know it exists)
    setup_identifier = prop.DefinitionRef(
        id="ce-setup-id",
        domain=prop.Domain.CustomEntity,
        scope="deployment",
        code="resource",
    )
    custom_inline = lumi.InlinePropertiesResource(
        id="test-custom-properties",
        provider="Lusid.CustomEntity.deployment",
        properties=[
            lumi.InlineProperty(
                key=setup_identifier,
                name="deployment_identifier",
                description="Identifier for an fbnconfig deployment resource",
            )
        ],
    )
    return [inline_properties, custom_inline]


def wait_for_it(client, sql, expected_rows):
    tries = 10
    sleep = 10
    while True:
        r = put_query(client, sql)
        if len(r) == expected_rows:
            return r
        tries -= 1
        if tries == 0:
            print("Failed to get", expected_rows, "after", tries, sql, r)
            assert len(r) == expected_rows, f"Failed to get {expected_rows}"
        time.sleep(sleep)


def test_teardown(setup_deployment, get_deployment):
    # Check if properties exist before creating
    check_path = (
        "config/lusid/factories/" f"{setup_deployment.name}_portfoliotransactionproviderfactory.csv"
    )
    check = f"select Content from sys.file where path = '{check_path}'"
    # Create the inline properties
    deployment = fbnconfig.Deployment(setup_deployment.name, get_deployment)
    fbnconfig.deploy(deployment, lusid_env, token)
    # Check it exists
    wait_for_it(client, check, 1)
    # Remove it
    fbnconfig.deploy(fbnconfig.Deployment(setup_deployment.name, []), lusid_env, token)
    # Check it's gone
    wait_for_it(client, check, 0)


def test_create(setup_deployment, get_deployment):
    deployment = fbnconfig.Deployment(setup_deployment.name, get_deployment)
    fbnconfig.deploy(deployment, lusid_env, token)
    check_path = (
        "config/lusid/factories/" f"{setup_deployment.name}_portfoliotransactionproviderfactory.csv"
    )
    r = put_query(client, f"select Content from sys.file where path = '{check_path}'")
    assert len(r) == 1
    # Verify the content contains our test property
    content = r[0]["Content"]
    assert "TestProperty" in content
    assert "Transaction/default/Custodian" in content


def test_update_with_extension_change(setup_deployment, get_deployment):
    # First deployment
    deployment = fbnconfig.Deployment(setup_deployment.name, get_deployment)
    fbnconfig.deploy(deployment, lusid_env, token)
    # Modify the provider_name_extension to trigger an update
    extension = "Ext" + setup_deployment.name
    get_deployment[0].provider_name_extension = extension
    update = fbnconfig.deploy(deployment, lusid_env, token)
    assert [a.change for a in update if a.type == "InlinePropertiesResource"] == ["update", "nochange"]
    # check the old one is gone
    old_path = (
        "config/lusid/factories/" f"{setup_deployment.name}_portfoliotransactionproviderfactory.csv"
    )
    old_check = f"select Content from sys.file where path = '{old_path}'"
    wait_for_it(client, old_check, 0)
    # check the new one has been created
    new_path = (
        "config/lusid/factories/" f"Ext{setup_deployment.name}_portfoliotransactionproviderfactory.csv"
    )
    check = f"select Content from sys.file where path = '{new_path}'"
    wait_for_it(client, check, 1)


def test_update_change(setup_deployment, get_deployment):
    # First deployment
    deployment = fbnconfig.Deployment(setup_deployment.name, get_deployment)
    fbnconfig.deploy(deployment, lusid_env, token)
    # Modify the provider_name_extension to trigger an update
    get_deployment[0].properties[0].description = "Updated description"
    update = fbnconfig.deploy(deployment, lusid_env, token)
    assert [a.change for a in update if a.type == "InlinePropertiesResource"] == ["update", "nochange"]


def test_nochange(setup_deployment, get_deployment):
    deployment = fbnconfig.Deployment(setup_deployment.name, get_deployment)
    fbnconfig.deploy(deployment, lusid_env, token)
    update = fbnconfig.deploy(deployment, lusid_env, token)
    assert [a.change for a in update if a.type == "InlinePropertiesResource"] == ["nochange", "nochange"]



import os
from types import SimpleNamespace

from httpx import HTTPStatusError
from pytest import fixture

import fbnconfig
from fbnconfig import notifications as notif
from tests.integration.generate_test_name import gen

if os.getenv("FBN_ACCESS_TOKEN") is None or os.getenv("LUSID_ENV") is None:
    raise RuntimeError(
        "Both FBN_ACCESS_TOKEN and LUSID_ENV variables need "
        "to be set to run integration tests"
    )


lusid_env = os.environ["LUSID_ENV"]
token = os.environ["FBN_ACCESS_TOKEN"]


@fixture(scope="module")
def setup_deployment():
    print("Creating new subscription")
    client = fbnconfig.create_client(lusid_env, token)

    # Deletes test subscription if it exsits
    try:
        client.delete("/notification/api/subscriptions/testScope/testCode")
    except Exception:
        pass

    # Creates new subscription to test notifications on
    desired = {
        "id": {
            "scope": "testScope",
            "code": "testCode"
        },
        "displayName": "TestDisplayName",
        "description": "TestDescription",
        "status": "Active",
        "matchingPattern": {
            "eventType": "Manual",
            "filter": "Body.Message eq 'TestMessage'"
        },
    }
    client.post("/notification/api/subscriptions", json=desired)

    deployment_name = gen("notifications")
    print(f"\nRunning for deployment {deployment_name}...")

    yield SimpleNamespace(name=deployment_name)
    # Teardown: Clean up resources (if any) after the test
    print("\nTearing down resources...")
    fbnconfig.deploy(fbnconfig.Deployment(deployment_name, []), lusid_env, token)
    client.delete("/notification/api/subscriptions/testScope/testCode")


def configure(env):
    """Create a test email notification."""
    deployment_name = getattr(env, "name", "notification")

    email_type = notif.EmailNotificationType(
        subject="Test Event: {{body.subject}}",
        plain_text_body="Event received: {{body.message}}",
        html_body="<p>Event received: <strong>{{body.message}}</strong></p>",
        email_address_to=["test@finbourne.com"],
        email_address_cc=["testcc@finbourne.com"]
    )

    subscription_ref = notif.SubscriptionRef(
        id="subscription_ref",
        scope="testScope",
        code="testCode"
    )

    email_notification = notif.NotificationResource(
        id="testEmailNotification",
        subscription=subscription_ref,
        notification_id=f"emailNotif001-{deployment_name}",
        display_name="Test Email Notification",
        description="Testing email notification creation",
        notification_type=email_type
    )

    matching_pattern = notif.MatchingPattern(
        event_type="Manual",
        filter="Body.Message eq 'Test'"
    )

    sub_resource = notif.SubscriptionResource(
        id="SubExampleId",
        scope=f"testScopeSub-{deployment_name}",
        code="testCodeSub",
        display_name="Example display name",
        description="Example description",
        status=notif.SubscriptionStatus.ACTIVE,
        matching_pattern=matching_pattern
    )

    sms_type = notif.SmsNotificationType(
        body="Example_body",
        recipients=["+47000000000"]
    )

    sms_notification = notif.NotificationResource(
        id="testSmsNotification",
        subscription=sub_resource,
        notification_id=f"smsNotif001-{deployment_name}",
        display_name="Test sms Notification",
        description="Testing sms notification creation",
        notification_type=sms_type
    )

    return fbnconfig.Deployment(deployment_name, [email_notification, sms_notification])


def test_teardown(setup_deployment):
    deployment_name = setup_deployment.name
    client = fbnconfig.create_client(lusid_env, token)
    # setup deployment
    fbnconfig.deploy(configure(setup_deployment), lusid_env, token)

    # check they exists
    client.request("get", "/notification/api/subscriptions/testScope/testCode/"
        f"notifications/emailNotif001-{deployment_name}")

    client.get(f"/notification/api/subscriptions/testScopeSub-{deployment_name}/testCodeSub/"
        f"notifications/smsNotif001-{deployment_name}")

    # Tear it down
    fbnconfig.deploy(fbnconfig.Deployment(deployment_name, []), lusid_env, token)

    # check they were deleted
    try:
        client.request("get", "/notification/api/subscriptions/testScope/testCode/"
        f"notifications/emailNotif001-{deployment_name}")
    except HTTPStatusError as error:
        assert error.response.status_code == 404

    try:
        client.get(f"/notification/api/subscriptions/testScopeSub-{deployment_name}/testCodeSub/"
        f"notifications/smsNotif001-{deployment_name}")
    except HTTPStatusError as error:
        assert error.response.status_code == 404


def test_create_email_notification(setup_deployment):
    deployment_name = setup_deployment.name
    fbnconfig.deploy(configure(setup_deployment), lusid_env, token)
    client = fbnconfig.create_client(lusid_env, token)

    # checks email and sms notifications are created
    email_response = client.get("/notification/api/subscriptions/testScope/testCode/"
    f"notifications/emailNotif001-{deployment_name}")

    sms_response = client.get(f"/notification/api/subscriptions/"
    f"testScopeSub-{deployment_name}/testCodeSub/notifications/smsNotif001-{deployment_name}")

    # assert codes
    assert email_response.status_code == 200
    assert sms_response.status_code == 200


def test_update_no_changes(setup_deployment):
    fbnconfig.deploy(configure(setup_deployment), lusid_env, token)
    update = fbnconfig.deploy(configure(setup_deployment), lusid_env, token)

    email_change = [a.change for a in update
                    if a.type == "NotificationResource" and a.id == "testEmailNotification"]
    sms_change = [a.change for a in update
                  if a.type == "NotificationResource" and a.id == "testSmsNotification"]
    assert email_change == ["nochange"]
    assert sms_change == ["nochange"]



from fbnconfig import Deployment, workspace

num_groups = 3
num_items = 2


def make_shared_branch(org, region):
    wksp = workspace.WorkspaceResource(
        id=f"wksp-{org}-{region}",
        visibility=workspace.Visibility.SHARED,
        name=f"{org}-{region}",
        description=f"workspace for {region} {org}",
    )
    items = [
        workspace.WorkspaceItemResource(
        id=f"item-{org}-{region}-{i}",
        workspace=wksp,
        group=f"group{1 + (i % num_groups)}",
        name=f"item{i}",
        description=f"item {i} version two",
        format=1,
        type="lusid-web-dashboard",
        content={"msg": "some text"},
    ) for i in range(0, num_items)]
    return [wksp] + items


def make_personal_branch(user):
    wksp = workspace.WorkspaceResource(
        id=f"wksp-personal-{user}",
        visibility=workspace.Visibility.PERSONAL,
        name=user,
        description=f"workspace for {user}",
    )
    num_groups = 2
    items = [
        workspace.WorkspaceItemResource(
        id=f"item-personal-{user}-{i}",
        workspace=wksp,
        group=f"group{1 + (i % num_groups)}",
        name=f"myitem{i}",
        description=f"item {i} version two",
        type="lusid-web-dashboard",
        format=1,
        content={"msg": "some text"},
    ) for i in range(0, num_items)]
    return [wksp] + items


def configure(env):
    deployment_name = getattr(env, "name", "workspace_example")
    orgs = ["operations", "equity-trading", "compliance", "risk-management", "market-data"]
    regions = ["emea", "apac", "us", "latam"]
    users = ["bill", "ted", "john", "paul", "ringo", "diana-ross", "florence", "mary-wilson",
        "micheal", "jermaine", "jackie", "marlon", "lionel", "william", "walter", "james-dean"]
    resources = []
    for org in orgs:
        for region in regions:
            resources.extend(make_shared_branch(org, region))
    for user in users:
        resources.extend(make_personal_branch(user))
    return Deployment(deployment_name, resources)
